/*
 * bsc1184952_fuse_file
 *
 * Fix for CVE-2020-36322, bsc#1184952 (fs/fuse/file.c part)
 *
 *  Copyright (c) 2021 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include "bsc1184952_common.h"

/* klp-ccp: from fs/fuse/fuse_i.h */
static void (*kgre_fuse_read_fill)(struct fuse_req *req, struct file *file,
		    loff_t pos, size_t count, int opcode);

int kgrp_fuse_open_common(struct inode *inode, struct file *file, bool isdir);

static struct fuse_file *(*kgre_fuse_file_get)(struct fuse_file *ff);

static void (*kgre_fuse_finish_open)(struct inode *inode, struct file *file);

int kgrp_fuse_fsync_common(struct file *file, loff_t start, loff_t end,
		      int datasync, int isdir);

static struct fuse_req *(*kgre_fuse_get_req)(struct fuse_conn *fc, unsigned npages);
static struct fuse_req *(*kgre_fuse_get_req_for_background)(struct fuse_conn *fc,
					     unsigned npages);

static struct fuse_req *(*kgre_fuse_get_req_nofail_nopages)(struct fuse_conn *fc,
					     struct file *file);

static void (*kgre_fuse_put_request)(struct fuse_conn *fc, struct fuse_req *req);

static void (*kgre_fuse_request_send)(struct fuse_conn *fc, struct fuse_req *req);

static ssize_t (*kgre_fuse_simple_request)(struct fuse_conn *fc, struct fuse_args *args);

static void (*kgre_fuse_request_send_background)(struct fuse_conn *fc, struct fuse_req *req);

static void (*kgre_fuse_invalidate_attr)(struct inode *inode);

static void (*kgre_fuse_invalidate_atime)(struct inode *inode);

static int (*kgre_fuse_allow_current_process)(struct fuse_conn *fc);

static u64 (*kgre_fuse_lock_owner_id)(struct fuse_conn *fc, fl_owner_t id);

static int (*kgre_fuse_update_attributes)(struct inode *inode, struct kstat *stat,
			   struct file *file, bool *refreshed);

static void (*kgre_fuse_set_nowrite)(struct inode *inode);
static void (*kgre_fuse_release_nowrite)(struct inode *inode);

static u64 (*kgre_fuse_get_attr_version)(struct fuse_conn *fc);

static int (*kgre_fuse_do_open)(struct fuse_conn *fc, u64 nodeid, struct file *file,
		 bool isdir);

static ssize_t (*kgre_fuse_direct_io)(struct fuse_io_priv *io, struct iov_iter *iter,
		       loff_t *ppos, int flags);
static long (*kgre_fuse_do_ioctl)(struct file *file, unsigned int cmd, unsigned long arg,
		   unsigned int flags);
long kgrp_fuse_ioctl_common(struct file *file, unsigned int cmd,
		       unsigned long arg, unsigned int flags);

static bool (*kgre_fuse_write_update_size)(struct inode *inode, loff_t pos);

/* klp-ccp: from fs/fuse/file.c */
#include <linux/pagemap.h>
#include <linux/slab.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/compat.h>
#include <linux/swap.h>
#include <linux/uio.h>

static void (*kgre_fuse_file_put)(struct fuse_file *ff, bool sync, bool isdir);

int kgrp_fuse_open_common(struct inode *inode, struct file *file, bool isdir)
{
	struct fuse_conn *fc = get_fuse_conn(inode);
	int err;
	bool lock_inode = (file->f_flags & O_TRUNC) &&
			  fc->atomic_o_trunc &&
			  fc->writeback_cache;

	/*
	 * Fix CVE-2020-36322
	 *  +3 lines
	 */
	if (kgrp_fuse_is_bad(inode))
		return -EIO;

	err = generic_file_open(inode, file);
	if (err)
		return err;

	if (lock_inode)
		mutex_lock(&inode->i_mutex);

	err = (*kgre_fuse_do_open)(fc, get_node_id(inode), file, isdir);

	if (!err)
		(*kgre_fuse_finish_open)(inode, file);

	if (lock_inode)
		mutex_unlock(&inode->i_mutex);

	return err;
}

static int (*kgre_fuse_wait_on_page_writeback)(struct inode *inode, pgoff_t index);

static void kgrr_fuse_sync_writes(struct inode *inode)
{
	(*kgre_fuse_set_nowrite)(inode);
	(*kgre_fuse_release_nowrite)(inode);
}

int kgrp_fuse_flush(struct file *file, fl_owner_t id)
{
	struct inode *inode = file_inode(file);
	struct fuse_conn *fc = get_fuse_conn(inode);
	struct fuse_file *ff = file->private_data;
	struct fuse_req *req;
	struct fuse_flush_in inarg;
	int err;

	/*
	 * Fix CVE-2020-36322
	 *  -1 line, +1 line
	 */
	if (kgrp_fuse_is_bad(inode))
		return -EIO;

	if (fc->no_flush)
		return 0;

	err = write_inode_now(inode, 1);
	if (err)
		return err;

	mutex_lock(&inode->i_mutex);
	kgrr_fuse_sync_writes(inode);
	mutex_unlock(&inode->i_mutex);

	if (test_bit(AS_ENOSPC, &file->f_mapping->flags) &&
	    test_and_clear_bit(AS_ENOSPC, &file->f_mapping->flags))
		err = -ENOSPC;
	if (test_bit(AS_EIO, &file->f_mapping->flags) &&
	    test_and_clear_bit(AS_EIO, &file->f_mapping->flags))
		err = -EIO;
	if (err)
		return err;

	req = (*kgre_fuse_get_req_nofail_nopages)(fc, file);
	memset(&inarg, 0, sizeof(inarg));
	inarg.fh = ff->fh;
	inarg.lock_owner = (*kgre_fuse_lock_owner_id)(fc, id);
	req->in.h.opcode = FUSE_FLUSH;
	req->in.h.nodeid = get_node_id(inode);
	req->in.numargs = 1;
	req->in.args[0].size = sizeof(inarg);
	req->in.args[0].value = &inarg;
	__set_bit(FR_FORCE, &req->flags);
	(*kgre_fuse_request_send)(fc, req);
	err = req->out.h.error;
	(*kgre_fuse_put_request)(fc, req);
	if (err == -ENOSYS) {
		fc->no_flush = 1;
		err = 0;
	}
	return err;
}

int kgrp_fuse_fsync_common(struct file *file, loff_t start, loff_t end,
		      int datasync, int isdir)
{
	struct inode *inode = file->f_mapping->host;
	struct fuse_conn *fc = get_fuse_conn(inode);
	struct fuse_file *ff = file->private_data;
	FUSE_ARGS(args);
	struct fuse_fsync_in inarg;
	int err;

	/*
	 * Fix CVE-2020-36322
	 *  -1 line, +1 line
	 */
	if (kgrp_fuse_is_bad(inode))
		return -EIO;

	mutex_lock(&inode->i_mutex);

	/*
	 * Start writeback against all dirty pages of the inode, then
	 * wait for all outstanding writes, before sending the FSYNC
	 * request.
	 */
	err = filemap_write_and_wait_range(inode->i_mapping, start, end);
	if (err)
		goto out;

	kgrr_fuse_sync_writes(inode);

	/*
	 * Due to implementation of fuse writeback
	 * filemap_write_and_wait_range() does not catch errors.
	 * We have to do this directly after fuse_sync_writes()
	 */
	if (test_bit(AS_ENOSPC, &file->f_mapping->flags) &&
	    test_and_clear_bit(AS_ENOSPC, &file->f_mapping->flags))
		err = -ENOSPC;
	if (test_bit(AS_EIO, &file->f_mapping->flags) &&
	    test_and_clear_bit(AS_EIO, &file->f_mapping->flags))
		err = -EIO;
	if (err)
		goto out;

	err = sync_inode_metadata(inode, 1);
	if (err)
		goto out;

	if ((!isdir && fc->no_fsync) || (isdir && fc->no_fsyncdir))
		goto out;

	memset(&inarg, 0, sizeof(inarg));
	inarg.fh = ff->fh;
	inarg.fsync_flags = datasync ? 1 : 0;
	args.in.h.opcode = isdir ? FUSE_FSYNCDIR : FUSE_FSYNC;
	args.in.h.nodeid = get_node_id(inode);
	args.in.numargs = 1;
	args.in.args[0].size = sizeof(inarg);
	args.in.args[0].value = &inarg;
	err = (*kgre_fuse_simple_request)(fc, &args);
	if (err == -ENOSYS) {
		if (isdir)
			fc->no_fsyncdir = 1;
		else
			fc->no_fsync = 1;
		err = 0;
	}
out:
	mutex_unlock(&inode->i_mutex);
	return err;
}

static int (*kgre_fuse_do_readpage)(struct file *file, struct page *page);

int kgrp_fuse_readpage(struct file *file, struct page *page)
{
	struct inode *inode = page->mapping->host;
	int err;

	err = -EIO;
	/*
	 * Fix CVE-2020-36322
	 *  -1 line, +1 line
	 */
	if (kgrp_fuse_is_bad(inode))
		goto out;

	err = (*kgre_fuse_do_readpage)(file, page);
	(*kgre_fuse_invalidate_atime)(inode);
 out:
	unlock_page(page);
	return err;
}

static void (*kgre_fuse_readpages_end)(struct fuse_conn *fc, struct fuse_req *req);

static void kgrr_fuse_send_readpages(struct fuse_req *req, struct file *file)
{
	struct fuse_file *ff = file->private_data;
	struct fuse_conn *fc = ff->fc;
	loff_t pos = page_offset(req->pages[0]);
	size_t count = req->num_pages << PAGE_CACHE_SHIFT;

	req->out.argpages = 1;
	req->out.page_zeroing = 1;
	req->out.page_replace = 1;
	(*kgre_fuse_read_fill)(req, file, pos, count, FUSE_READ);
	req->misc.read.attr_ver = (*kgre_fuse_get_attr_version)(fc);
	if (fc->async_read) {
		req->ff = (*kgre_fuse_file_get)(ff);
		req->end = (*kgre_fuse_readpages_end);
		(*kgre_fuse_request_send_background)(fc, req);
	} else {
		(*kgre_fuse_request_send)(fc, req);
		(*kgre_fuse_readpages_end)(fc, req);
		(*kgre_fuse_put_request)(fc, req);
	}
}

struct fuse_fill_data {
	struct fuse_req *req;
	struct file *file;
	struct inode *inode;
	unsigned nr_pages;
};

static int (*kgre_fuse_readpages_fill)(void *_data, struct page *page);

int kgrp_fuse_readpages(struct file *file, struct address_space *mapping,
			  struct list_head *pages, unsigned nr_pages)
{
	struct inode *inode = mapping->host;
	struct fuse_conn *fc = get_fuse_conn(inode);
	struct fuse_fill_data data;
	int err;
	int nr_alloc = min_t(unsigned, nr_pages, FUSE_MAX_PAGES_PER_REQ);

	err = -EIO;
	/*
	 * Fix CVE-2020-36322
	 *  -1 line, +1 line
	 */
	if (kgrp_fuse_is_bad(inode))
		goto out;

	data.file = file;
	data.inode = inode;
	if (fc->async_read)
		data.req = (*kgre_fuse_get_req_for_background)(fc, nr_alloc);
	else
		data.req = (*kgre_fuse_get_req)(fc, nr_alloc);
	data.nr_pages = nr_pages;
	err = PTR_ERR(data.req);
	if (IS_ERR(data.req))
		goto out;

	err = read_cache_pages(mapping, pages, (*kgre_fuse_readpages_fill), &data);
	if (!err) {
		if (data.req->num_pages)
			kgrr_fuse_send_readpages(data.req, file);
		else
			(*kgre_fuse_put_request)(fc, data.req);
	}
out:
	return err;
}

static size_t (*kgre_fuse_send_write)(struct fuse_req *req, struct fuse_io_priv *io,
			      loff_t pos, size_t count, fl_owner_t owner);

static size_t kgrr_fuse_send_write_pages(struct fuse_req *req, struct file *file,
				    struct inode *inode, loff_t pos,
				    size_t count)
{
	size_t res;
	unsigned offset;
	unsigned i;
	struct fuse_io_priv io = FUSE_IO_PRIV_SYNC(file);

	for (i = 0; i < req->num_pages; i++)
		(*kgre_fuse_wait_on_page_writeback)(inode, req->pages[i]->index);

	res = (*kgre_fuse_send_write)(req, &io, pos, count, NULL);

	offset = req->page_descs[0].offset;
	count = res;
	for (i = 0; i < req->num_pages; i++) {
		struct page *page = req->pages[i];

		if (!req->out.h.error && !offset && count >= PAGE_CACHE_SIZE)
			SetPageUptodate(page);

		if (count > PAGE_CACHE_SIZE - offset)
			count -= PAGE_CACHE_SIZE - offset;
		else
			count = 0;
		offset = 0;

		unlock_page(page);
		page_cache_release(page);
	}

	return res;
}

static ssize_t fuse_fill_write_pages(struct fuse_req *req,
			       struct address_space *mapping,
			       struct iov_iter *ii, loff_t pos)
{
	struct fuse_conn *fc = get_fuse_conn(mapping->host);
	unsigned offset = pos & (PAGE_CACHE_SIZE - 1);
	size_t count = 0;
	int err;

	req->in.argpages = 1;
	req->page_descs[0].offset = offset;

	do {
		size_t tmp;
		struct page *page;
		pgoff_t index = pos >> PAGE_CACHE_SHIFT;
		size_t bytes = min_t(size_t, PAGE_CACHE_SIZE - offset,
				     iov_iter_count(ii));

		bytes = min_t(size_t, bytes, fc->max_write - count);

 again:
		err = -EFAULT;
		if (iov_iter_fault_in_readable(ii, bytes))
			break;

		err = -ENOMEM;
		page = grab_cache_page_write_begin(mapping, index, 0);
		if (!page)
			break;

		if (mapping_writably_mapped(mapping))
			flush_dcache_page(page);

		tmp = iov_iter_copy_from_user_atomic(page, ii, offset, bytes);
		flush_dcache_page(page);

		iov_iter_advance(ii, tmp);
		if (!tmp) {
			unlock_page(page);
			page_cache_release(page);
			bytes = min(bytes, iov_iter_single_seg_count(ii));
			goto again;
		}

		err = 0;
		req->pages[req->num_pages] = page;
		req->page_descs[req->num_pages].length = tmp;
		req->num_pages++;

		count += tmp;
		pos += tmp;
		offset += tmp;
		if (offset == PAGE_CACHE_SIZE)
			offset = 0;

		if (!fc->big_writes)
			break;
	} while (iov_iter_count(ii) && count < fc->max_write &&
		 req->num_pages < req->max_pages && offset == 0);

	return count > 0 ? count : err;
}

static inline unsigned fuse_wr_pages(loff_t pos, size_t len)
{
	return min_t(unsigned,
		     ((pos + len - 1) >> PAGE_CACHE_SHIFT) -
		     (pos >> PAGE_CACHE_SHIFT) + 1,
		     FUSE_MAX_PAGES_PER_REQ);
}

ssize_t kgrp_fuse_perform_write(struct file *file,
				  struct address_space *mapping,
				  struct iov_iter *ii, loff_t pos)
{
	struct inode *inode = mapping->host;
	struct fuse_conn *fc = get_fuse_conn(inode);
	struct fuse_inode *fi = get_fuse_inode(inode);
	int err = 0;
	ssize_t res = 0;

	/*
	 * Fix CVE-2020-36322
	 *  -1 line, +1 line
	 */
	if (kgrp_fuse_is_bad(inode))
		return -EIO;

	if (inode->i_size < pos + iov_iter_count(ii))
		set_bit(FUSE_I_SIZE_UNSTABLE, &fi->state);

	do {
		struct fuse_req *req;
		ssize_t count;
		unsigned nr_pages = fuse_wr_pages(pos, iov_iter_count(ii));

		req = (*kgre_fuse_get_req)(fc, nr_pages);
		if (IS_ERR(req)) {
			err = PTR_ERR(req);
			break;
		}

		count = fuse_fill_write_pages(req, mapping, ii, pos);
		if (count <= 0) {
			err = count;
		} else {
			size_t num_written;

			num_written = kgrr_fuse_send_write_pages(req, file, inode,
							    pos, count);
			err = req->out.h.error;
			if (!err) {
				res += num_written;
				pos += num_written;

				/* break out of the loop on short write */
				if (num_written != count)
					err = -EIO;
			}
		}
		(*kgre_fuse_put_request)(fc, req);
	} while (!err && iov_iter_count(ii));

	if (res > 0)
		(*kgre_fuse_write_update_size)(inode, pos);

	clear_bit(FUSE_I_SIZE_UNSTABLE, &fi->state);
	(*kgre_fuse_invalidate_attr)(inode);

	return res > 0 ? res : err;
}

ssize_t kgrp_fuse_file_write_iter(struct kiocb *iocb, struct iov_iter *from)
{
	struct file *file = iocb->ki_filp;
	struct address_space *mapping = file->f_mapping;
	ssize_t written = 0;
	ssize_t written_buffered = 0;
	struct inode *inode = mapping->host;
	ssize_t err;
	loff_t endbyte = 0;

	/*
	 * Fix CVE-2020-36322
	 *  +3 lines
	 */
	if (kgrp_fuse_is_bad(inode))
		return -EIO;

	if (get_fuse_conn(inode)->writeback_cache) {
		/* Update size (EOF optimization) and mode (SUID clearing) */
		err = (*kgre_fuse_update_attributes)(mapping->host, NULL, file, NULL);
		if (err)
			return err;

		return generic_file_write_iter(iocb, from);
	}

	mutex_lock(&inode->i_mutex);

	/* We can write back this queue in page reclaim */
	current->backing_dev_info = inode_to_bdi(inode);

	err = generic_write_checks(iocb, from);
	if (err <= 0)
		goto out;

	err = file_remove_privs(file);
	if (err)
		goto out;

	err = file_update_time(file);
	if (err)
		goto out;

	if (iocb->ki_flags & IOCB_DIRECT) {
		loff_t pos = iocb->ki_pos;
		written = generic_file_direct_write(iocb, from);
		if (written < 0 || !iov_iter_count(from))
			goto out;

		pos += written;

		written_buffered = kgrp_fuse_perform_write(file, mapping, from, pos);
		if (written_buffered < 0) {
			err = written_buffered;
			goto out;
		}
		endbyte = pos + written_buffered - 1;

		err = filemap_write_and_wait_range(file->f_mapping, pos,
						   endbyte);
		if (err)
			goto out;

		invalidate_mapping_pages(file->f_mapping,
					 pos >> PAGE_CACHE_SHIFT,
					 endbyte >> PAGE_CACHE_SHIFT);

		written += written_buffered;
		iocb->ki_pos = pos + written_buffered;
	} else {
		written = kgrp_fuse_perform_write(file, mapping, from, iocb->ki_pos);
		if (written >= 0)
			iocb->ki_pos += written;
	}
out:
	current->backing_dev_info = NULL;
	mutex_unlock(&inode->i_mutex);

	return written ? written : err;
}

ssize_t kgrp___fuse_direct_read(struct fuse_io_priv *io,
				  struct iov_iter *iter,
				  loff_t *ppos)
{
	ssize_t res;
	struct file *file = io->file;
	struct inode *inode = file_inode(file);

	/*
	 * Fix CVE-2020-36322
	 *  -1 line, +1 line
	 */
	if (kgrp_fuse_is_bad(inode))
		return -EIO;

	res = (*kgre_fuse_direct_io)(io, iter, ppos, 0);

	(*kgre_fuse_invalidate_attr)(inode);

	return res;
}

ssize_t kgrp_fuse_direct_write_iter(struct kiocb *iocb, struct iov_iter *from)
{
	struct file *file = iocb->ki_filp;
	struct inode *inode = file_inode(file);
	struct fuse_io_priv io = FUSE_IO_PRIV_SYNC(file);
	ssize_t res;

	/*
	 * Fix CVE-2020-36322
	 *  -1 line, +1 line
	 */
	if (kgrp_fuse_is_bad(inode))
		return -EIO;

	/* Don't allow parallel writes to the same file */
	mutex_lock(&inode->i_mutex);
	res = generic_write_checks(iocb, from);
	if (res > 0)
		res = (*kgre_fuse_direct_io)(&io, from, &iocb->ki_pos, FUSE_DIO_WRITE);
	(*kgre_fuse_invalidate_attr)(inode);
	if (res > 0)
		(*kgre_fuse_write_update_size)(inode, iocb->ki_pos);
	mutex_unlock(&inode->i_mutex);

	return res;
}

struct fuse_fill_wb_data {
	struct fuse_req *req;
	struct fuse_file *ff;
	struct inode *inode;
	struct page **orig_pages;
};

static void (*kgre_fuse_writepages_send)(struct fuse_fill_wb_data *data);

static int (*kgre_fuse_writepages_fill)(struct page *page,
		struct writeback_control *wbc, void *_data);

int kgrp_fuse_writepages(struct address_space *mapping,
			   struct writeback_control *wbc)
{
	struct inode *inode = mapping->host;
	struct fuse_fill_wb_data data;
	int err;

	err = -EIO;
	/*
	 * Fix CVE-2020-36322
	 *  -1 line, +1 line
	 */
	if (kgrp_fuse_is_bad(inode))
		goto out;

	data.inode = inode;
	data.req = NULL;
	data.ff = NULL;

	err = -ENOMEM;
	data.orig_pages = kcalloc(FUSE_MAX_PAGES_PER_REQ,
				  sizeof(struct page *),
				  GFP_NOFS);
	if (!data.orig_pages)
		goto out;

	err = write_cache_pages(mapping, wbc, (*kgre_fuse_writepages_fill), &data);
	if (data.req) {
		/* Ignore errors if we can write at least one page */
		BUG_ON(!data.req->num_pages);
		(*kgre_fuse_writepages_send)(&data);
		err = 0;
	}
	if (data.ff)
		(*kgre_fuse_file_put)(data.ff, false, false);

	kfree(data.orig_pages);
out:
	return err;
}

long kgrp_fuse_ioctl_common(struct file *file, unsigned int cmd,
		       unsigned long arg, unsigned int flags)
{
	struct inode *inode = file_inode(file);
	struct fuse_conn *fc = get_fuse_conn(inode);

	if (!(*kgre_fuse_allow_current_process)(fc))
		return -EACCES;

	/*
	 * Fix CVE-2020-36322
	 *  -1 line, +1 line
	 */
	if (kgrp_fuse_is_bad(inode))
		return -EIO;

	return (*kgre_fuse_do_ioctl)(file, cmd, arg, flags);
}



#include <linux/kernel.h>
#include <linux/module.h>
#include "kgr_patch_bsc1184952.h"
#include "../kallsyms_relocs.h"

#define KGR_PATCHED_MODULE "fuse"

static struct kgr_kallsyms_reloc kgr_funcs[] = {
	{ "fuse_writepages_send", (void *)&kgre_fuse_writepages_send, "fuse" },
	{ "fuse_writepages_fill", (void *)&kgre_fuse_writepages_fill, "fuse" },
	{ "fuse_read_fill", (void *)&kgre_fuse_read_fill, "fuse" },
	{ "fuse_file_get", (void *)&kgre_fuse_file_get, "fuse" },
	{ "fuse_finish_open", (void *)&kgre_fuse_finish_open, "fuse" },
	{ "fuse_get_req", (void *)&kgre_fuse_get_req, "fuse" },
	{ "fuse_get_req_for_background",
	  (void *)&kgre_fuse_get_req_for_background, "fuse" },
	{ "fuse_get_req_nofail_nopages",
	  (void *)&kgre_fuse_get_req_nofail_nopages, "fuse" },
	{ "fuse_put_request", (void *)&kgre_fuse_put_request, "fuse" },
	{ "fuse_request_send", (void *)&kgre_fuse_request_send, "fuse" },
	{ "fuse_simple_request", (void *)&kgre_fuse_simple_request, "fuse" },
	{ "fuse_request_send_background",
	  (void *)&kgre_fuse_request_send_background, "fuse" },
	{ "fuse_invalidate_attr", (void *)&kgre_fuse_invalidate_attr,
	  "fuse" },
	{ "fuse_invalidate_atime", (void *)&kgre_fuse_invalidate_atime,
	  "fuse" },
	{ "fuse_allow_current_process",
	  (void *)&kgre_fuse_allow_current_process, "fuse" },
	{ "fuse_lock_owner_id", (void *)&kgre_fuse_lock_owner_id, "fuse" },
	{ "fuse_update_attributes", (void *)&kgre_fuse_update_attributes,
	  "fuse" },
	{ "fuse_set_nowrite", (void *)&kgre_fuse_set_nowrite, "fuse" },
	{ "fuse_release_nowrite", (void *)&kgre_fuse_release_nowrite, "fuse" },
	{ "fuse_get_attr_version", (void *)&kgre_fuse_get_attr_version,
	  "fuse" },
	{ "fuse_do_open", (void *)&kgre_fuse_do_open, "fuse" },
	{ "fuse_direct_io", (void *)&kgre_fuse_direct_io, "fuse" },
	{ "fuse_do_ioctl", (void *)&kgre_fuse_do_ioctl, "fuse" },
	{ "fuse_write_update_size", (void *)&kgre_fuse_write_update_size,
	  "fuse" },
	{ "fuse_wait_on_page_writeback",
	  (void *)&kgre_fuse_wait_on_page_writeback, "fuse" },
	{ "fuse_file_put", (void *)&kgre_fuse_file_put, "fuse" },
	{ "fuse_do_readpage", (void *)&kgre_fuse_do_readpage, "fuse" },
	{ "fuse_readpages_end", (void *)&kgre_fuse_readpages_end, "fuse" },
	{ "fuse_readpages_fill", (void *)&kgre_fuse_readpages_fill, "fuse" },
	{ "fuse_send_write", (void *)&kgre_fuse_send_write, "fuse" },
};

static int kgr_patch_bsc1184952_module_notify(struct notifier_block *nb,
					      unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = __kgr_resolve_kallsyms_relocs(kgr_funcs, ARRAY_SIZE(kgr_funcs));
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc1184952_module_nb = {
	.notifier_call = kgr_patch_bsc1184952_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc1184952_fuse_file_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = __kgr_resolve_kallsyms_relocs(kgr_funcs,
						    ARRAY_SIZE(kgr_funcs));
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc1184952_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc1184952_fuse_file_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1184952_module_nb);
}
